{$A-,B-,D-,E-,F-,G-,I+,L-,N-,O-,R-,S-,V-,X+}
(*
   CSTRING.PAS
   A "C-style" library for strings.

   Copyright (c) 1992,93 Alessandro Scotti
*)
unit CString;
interface

function StrChr( S: string; C: char ): byte;
function StrCmp( S1, S2: string ): integer;
function StrIChr( S: string; C: char ): byte;
function StrICmp( S1, S2: string ): integer;
function StrLwr( S: string ): string;
function StrRChr( S: string; C: char ): byte;
function StrRev( S: string ): string;
function StrTok( S1, S2: string ): string;
function StrUpr( S: string ): string;

function StrLft( S: string; N: byte ): string;
function StrOf( C: char; N: byte ): string;
function StrRgt( S: string; N: byte ): string;
function StrLTrim( S: string ): string;
function StrRTrim( S: string ): string;
function StrLPad( S: string; Len: byte ): string;
function StrRPad( S: string; Len: byte ): string;

function StrC2TP( S: pointer ): string;
function StrTP2C( S: string; D: pointer ): pointer;

function IToA( N: integer; Radix: byte ): string;
function WToA( N: word; Radix: byte ): string;
function LToH( N: longint ): string;
function PToH( P: pointer ): string;
function WToH( N: word ): string;

function MemChr( S: pointer; C: char; N: word ): pointer;
function MemCmp( S1, S2: pointer; N: word ): integer;
function MemPos( S: pointer; C: char; N: word ): word;
function MemSet( S: pointer; C: byte; N: word ): pointer;
function Mem2Str( S: pointer; N: byte ): string;

implementation

(*
   Finds the first occurence of C in S.

   Output:
          the position of C in S, or 0 if S does not contain C.
*)
function StrChr( S: string; C: char ): byte; assembler;
asm
  mov   al, [C]
  les   di, [S]
  cld
  mov   cl, es:[di]                     (* Get string length *)
  xor   ch, ch
  inc   di
  mov   si, di
  jcxz  @@Exit                          (* Exit if null string *)
  repne scasb                           (* Scan the string *)
  je    @@Exit                          (* Exit if found *)
  mov   di, si
@@Exit:
  sub   di, si                          (* Get position of character... *)
  mov   ax, di                          (* ...and return it *)
end;

(*
   Finds the first occurence of C in S, ignoring case.

   Output:
          the position of C in S, or 0 if S does not contain C.
*)
function StrIChr( S: string; C: char ): byte; assembler;
asm
  push  ds
  mov   ah, [C]
  lds   si, [S]
  cld
  mov   cl, ds:[si]                     (* Get string length *)
  xor   ch, ch
  inc   si
  mov   di, si
  jcxz  @@Exit                          (* Exit if null string *)
  cmp   ah, 'a'                         (* Convert C to uppercase *)
  jb    @@Loop
  cmp   ah, 'z'
  ja    @@Loop
  sub   ah, 'a'-'A'
@@Loop:
  lodsb                                 (* Get string char *)
  cmp   al, 'a'                         (* Convert it to uppercase *)
  jb    @@Test
  cmp   al, 'z'
  ja    @@Test
  sub   al, 'a'-'A'
@@Test:
  cmp   al, ah                          (* Do characters match? *)
  je    @@Exit                          (* Yes, exit *)
  loop  @@Loop
  mov   di, si
@@Exit:
  sub   si, di                          (* Get position of character... *)
  mov   ax, si                          (* ...and return it *)
  pop   ds
end;

(*
   Compares one string to another.

   Output:
          -1    if S1 is less than S2;
           0    if S1 is the same as S2;
           1    if S1 is greater than S2.
*)
function StrCmp( S1, S2: string ): integer; assembler;
asm
  mov   dx, ds                          (* Save DS *)
  cld
  lds   si, [S1]
  les   di, [S2]
  lodsb                                 (* Put length of S1 in AL *)
  mov   ah, es:[di]                     (* Put length of S2 in AL *)
  inc   di
  mov   cl, al
  cmp   cl, ah                          (* Is S1 shorter than S2? *)
  jb    @@Continue                      (* Yes, continue *)
  mov   cl, ah                          (* Loop on shortest length *)
@@Continue:
  xor   ch, ch
  jcxz  @@TestLen                       (* If one string is empty test lengths *)
  repe  cmpsb                           (* Compare string characters *)
  jne   @@Done                          (* Exit if they don't match... *)
@@TestLen:
  cmp   al, ah                          (* ...otherwise compare lengths *)
@@Done:
  mov   al, -1                          (* Assume S1 < S2, don't destroy flags! *)
  jb    @@Exit
  mov   al, +1                          (* Assume S1 > S2, don't destroy flags! *)
  ja    @@Exit
  xor   al, al                          (* S1 = S2 *)
@@Exit:
  cbw                                   (* Convert it to an integer *)
  mov   ds, dx                          (* Restore DS *)
end;

(*
   Compares one string to another, ignoring case.

   Output:
          -1    if S1 is less than S2;
           0    if S1 is the same as S2;
           1    if S1 is greater than S2.
*)
function StrICmp( S1, S2: string ): integer; assembler;
asm
  mov   dx, ds                          (* Save DS *)
  cld
  lds   si, [S1]
  les   di, [S2]
  lodsb                                 (* Put length of S1 in AL *)
  mov   ah, es:[di]                     (* Put length of S2 in AL *)
  inc   di
  mov   cl, al
  cmp   cl, ah                          (* Is S1 shorter than S2? *)
  jb    @@Continue                      (* Yes, continue *)
  mov   cl, ah                          (* Loop on shortest length *)
@@Continue:
  mov   bx, ax
  xor   ch, ch
  jcxz  @@TestLen                       (* If one string is empty test lengths *)
@@Rep:
  lodsb
  mov   ah, es:[di]
  inc   di
  cmp   al, 'A'                         (* Letter? *)
  jb    @@Cmp                           (* No, compare chars *)
  cmp   ah, 'A'                         (* Letter? *)
  jb    @@Cmp                           (* No, compare chars *)
  cmp   al, 'z'                         (* Letter? *)
  ja    @@Cmp                           (* No, compare chars *)
  cmp   ah, 'z'                         (* Letter? *)
  ja    @@Cmp                           (* No, compare chars *)
(*
   If we are here then AL and AH are probably letters. However,
   there are a few ASCII codes between 90 ('Z') and 97 ('a') and
   both AL and AH maybe one of these. When we set the 5th bit of
   AL and AH lowercase letters are converted to uppercase, which
   is exactly what we need, and extra codes fall in the range
   123..128 (91+32..96+32). But both AL and AH are below 122 ('z')
   at this point, so there are no side effects.
   This little "trick" should improve speed, at least when AL and
   AH are not both letters.
*)
  or    ax, $2020
@@Cmp:
  cmp   al, ah
  loope @@Rep
  jne   @@Done                          (* Exit if they don't match... *)
@@TestLen:
  cmp   bl, bh                          (* ...otherwise compare lengths *)
@@Done:
  mov   al, -1                          (* Assume S1 < S2, don't destroy flags! *)
  jb    @@Exit
  mov   al, +1                          (* Assume S1 > S2, don't destroy flags! *)
  ja    @@Exit
  xor   al, al                          (* S1 = S2 *)
@@Exit:
  cbw                                   (* Convert it to an integer *)
  mov   ds, dx                          (* Restore DS *)
end;


(*
   Converts S to lowercase.

   Output:
          converted string.
*)
function StrLwr( S: string ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  cld
  lodsb
  stosb
  cbw
  mov   cx, ax
  jcxz  @@Exit
@@Loop:
  lodsb
  cmp   al, 'A'
  jb    @@PutChar
  cmp   al, 'Z'
  ja    @@PutChar
  add   al, 'a'-'A'
@@PutChar:
  stosb
  loop  @@Loop
@@Exit:
  pop   ds
end;

(*
   Converts S to uppercase.

   Output:
          converted string.
*)
function StrUpr( S: string ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  cld
  lodsb
  stosb
  cbw
  mov   cx, ax
  jcxz  @@Exit
@@Loop:
  lodsb
  cmp   al, 'a'
  jb    @@PutChar
  cmp   al, 'z'
  ja    @@PutChar
  sub   al, 'a'-'A'
@@PutChar:
  stosb
  loop  @@Loop
@@Exit:
  pop   ds
end;

(*
   Finds the last occurence of C in S.

   Output:
          the position of C in S, or 0 if S does not contain C.
*)
function StrRChr( S: string; C: char ): byte; assembler;
asm
  mov   al, [C]
  les   di, [S]
  mov   cl, es:[di]
  xor   ch, ch
  mov   si, di                          (* Save string offset *)
  jcxz  @@Exit                          (* Exit if empty string *)
  add   di, cx                          (* Move to last char *)
  dec   si
  std                                   (* Move backward *)
  repne scasb                           (* Search the char *)
  je    @@Exit                          (* Exit if found *)
  mov   di, si                          (* Reset string pointer *)
@@Exit:
  sub   di, si
  mov   ax, di
end;

(*
   Reverses all characters in S.

   Output:
          the reversed string.
*)
function StrRev( S: string ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  cld
  lodsb
  stosb
  xor   ah, ah
  mov   cx, ax
  jcxz  @@Exit
  add   di, cx
@@Loop:
  lodsb
  dec   di
  mov   es:[di], al
  loop  @@Loop
@@Exit:
  pop   ds
end;

(*
   Scans S1 for the first token not contained in S2.

   Output:
          the first token found (an empty string if no tokens are found).

   Note:
          subsequent calls to StrTok with S1 empty use the previous S1
          string starting after the last token found.

   Example:
          WriteLn( StrTok( 'TEST /1 /2', ' /' ) );
          WriteLn( StrTok( '',           ' /' ) );
          WriteLn( StrTok( '',           ' /' ) );
          WriteLn( StrTok( '',           ' /' ) );
*)
function StrTok( S1, S2: string ): string;
const
  S: string = '';
var
  I1: byte;
  L1: byte absolute S1;
  L2: byte absolute S2;
begin
  StrTok := '';
  if( L2 = 0 ) then Exit;

  S2 := S2 + #00;
  if( S1 = '' ) then
    S1 := S
  else S1 := S1 + #00;

  while( L1 > 0 )and( Pos( S1[1], S2 ) > 0 ) do
    Delete( S1, 1, 1 );

  if( L1 > 0 ) then begin
    I1 := 1;
    while( Pos( S1[I1], S2 ) = 0 ) do
      Inc( I1 );
    StrTok := Copy( S1, 1, I1-1 );
    S := Copy( S1, I1, 255 );
  end;
end;

(*
   Unsigned (word) to ASCII.
*)
function WToA( N: word; Radix: byte ): string; assembler;
asm
  les   di, @Result
  mov   BYTE PTR es:[di], $00           (* Clear string length *)
  mov   ax, [N]
  mov   bl, [Radix]
  cmp   bl, $02                         (* Is the radix less than 2? *)
  jb    @@Exit                          (* Yes, exit *)
  xor   bh, bh
  xor   cx, cx
@@Loop1:
  xor   dx, dx
  div   bx                              (* Divide the number by radix *)
  push  dx                              (* Save remainder *)
  inc   cx                              (* Keep count of pushed words *)
  or    ax, ax                          (* Is the quotient zero? *)
  jnz   @@Loop1                         (* No, keep looping *)
  mov   es:[di], cl                     (* Update string length *)
  inc   di
@@Loop2:
  pop   ax                              (* Get last result *)
  add   al, '0'                         (* Convert remainder to digit *)
  cmp   al, '9'
  jbe   @@Store
  add   al, 'A'-'0'-$0A
@@Store:
  mov   es:[di], al                     (* Store digit into string *)
  inc   di
  loop  @@Loop2                         (* Loop for all digits *)
@@Exit:
end;

(*
   Integer to ASCII.
*)
function IToA( N: integer; Radix: byte ): string; assembler;
asm
  les   di, @Result
  mov   si, di
  inc   di
  mov   BYTE PTR es:[si], $00
  mov   ax, [N]
  mov   bl, [Radix]
  cmp   bl, $02                         (* Is the radix less than 2? *)
  jb    @@Exit                          (* Yes, exit *)
  xor   bh, bh
  xor   cx, cx
  or    ax, ax                          (* Is number non-negative? *)
  jns   @@Loop1                         (* Yes, continue *)
  inc   BYTE PTR es:[si]                (* Increment string length *)
  mov   BYTE PTR es:[di], '-'           (* Store a minus sign *)
  inc   di
  neg   ax                              (* Take absolute value of number *)
@@Loop1:
  xor   dx, dx
  div   bx
  push  dx                              (* Save remainder *)
  inc   cx
  or    ax, ax
  jnz   @@Loop1
  add   es:[si], cl
@@Loop2:
  pop   ax
  add   al, '0'
  cmp   al, '9'
  jbe   @@Store
  add   al, 'A'-'0'-$0A
@@Store:
  mov   es:[di], al
  inc   di
  loop  @@Loop2
@@Exit:
end;

(*
   Unsigned (word) to four-digit hexadecimal.
*)
function WToH( N: word ): string; assembler;
asm
  les   di, @Result
  cld
  mov   cx, $0404
  mov   es:[di], cl
  inc   di
  mov   ax, [N]
@@Loop:
  rol   ax, cl
  push  ax
  and   al, $0F
  add   al, '0'
  cmp   al, '9'
  jbe   @@Store
  add   al, 'A'-'0'-$0A
@@Store:
  stosb
  pop   ax
  dec   ch
  jnz   @@Loop
end;

(*
   Long to eight-digit hexadecimal.
*)
function LToH( N: longint ): string;
type
  Cast = record
    Lo, Hi: word;
  end;
var
  C: Cast absolute N;
begin
  LToH := WToH( C.Hi ) + WToH( C.Lo );
end;

(*
   Pointer to hexadecimal segment:offset.
*)
function PToH( P: pointer ): string;
type
  Cast = record
    Lo, Hi: word;
  end;
var
  C: Cast absolute P;
begin
  PToH := WToH( C.Hi ) + ':' + WToH( C.Lo );
end;

(*
   Searches the first N bytes of array S for character C.

   Output:
          a pointer to the first occurrence of C in S^ or NIL
          if C was not found.
*)
function MemChr( S: pointer; C: char; N: word ): pointer; assembler;
asm
  xor   dx, dx
  xor   bx, bx
  les   di, [S]
  mov   al, [C]
  mov   cx, [N]
  cld
  repne scasb
  jne   @@Exit
  dec   di
  mov   bx, di
  mov   dx, es
@@Exit:
  mov   ax, bx
end;

(*
   Searches the first N bytes of array S for character C.

   Output:
          the 1-based index to the first occurrence of C in S^ or 0
          if C was not found.
*)
function MemPos( S: pointer; C: char; N: word ): word; assembler;
asm
  les   di, [S]
  mov   dx, di
  mov   al, [C]
  mov   cx, [N]
  cld
  repne scasb
  mov   ax, 0
  jne   @@Exit
  mov   ax, di
  sub   ax, dx
@@Exit:
end;

(*
   Sets the first N bytes of S to C.

   Output:
          returns S.
*)
function MemSet( S: pointer; C: byte; N: word ): pointer; assembler;
asm
  les   di, [S]
  mov   bx, di
  mov   cx, [N]
  mov   al, [C]
  cld
  shr   cx, 1                           (* Get number of words to set... *)
  rep   stosw                           (* ...and set them *)
  rcl   cx, 1                           (* Restore carry *)
  rep   stosb                           (* Set the "odd" byte if any *)
  mov   ax, bx
  mov   dx, es
end;

(*
   Compares the first N bytes of S1 and S2.

   Output:
          -1    if S1 is less than S2;
           0    if S1 is the same as S2;
           1    if S1 is greater than S2.
*)
function MemCmp( S1, S2: pointer; N: word ): integer; assembler;
asm
  mov   dx, ds                          (* Save DS *)
  cld
  lds   si, [S1]
  les   di, [S2]
  mov   cx, [N]
  repe  cmpsb                           (* Compare string characters *)
  mov   al, -1                          (* Assume S1 < S2, don't trash flags! *)
  jb    @@Exit
  mov   al, +1                          (* Assume S1 > S2, don't trash flags! *)
  ja    @@Exit
  xor   al, al                          (* S1 = S2 *)
@@Exit:
  cbw                                   (* Convert it to an integer *)
  mov   ds, dx                          (* Restore DS *)
end;

(*
   Returns the first N characters in S.
*)
function StrLft( S: string; N: byte ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  cld
  lodsb
  cmp   al, [N]
  jbe   @@Copy
  mov   al, [N]
@@Copy:
  stosb
  mov   cl, al
  xor   ch, ch
  rep   movsb
  pop   ds
end;

(*
   Returns the last N characters in S.
*)
function StrRgt( S: string; N: byte ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  cld
  lodsb
  cmp   al, [N]
  jbe   @@Copy
  sub   al, [N]
  cbw
  add   si, ax
  mov   al, [N]
@@Copy:
  stosb
  mov   cl, al
  xor   ch, ch
  rep   movsb
  pop   ds
end;

(*
   Returns a string made of N characters C.
*)
function StrOf( C: char; N: byte ): string; assembler;
asm
  les   di, @Result
  cld
  mov   al, [N]
  stosb
  cbw
  xchg  cx, ax                          (* One byte shorter than "mov"... *)
  mov   al, [C]
  rep   stosb
end;

(*
   Trims leading spaces out of string S.
*)
function StrLTrim( S: string ): string; assembler;
asm
  push  ds
  lds   si, [S]
  cld
  lodsb
  mov   cl, al
  xor   ch, ch
  jcxz  @@Exit
@@Loop:
  cmp   BYTE PTR ds:[si], ' '
  jne   @@Exit
  inc   si
  loop  @@Loop
@@Exit:
  les   di, @Result
  mov   es:[di], cl
  inc   di
  rep   movsb
  pop   ds
end;

(*
   Trims trailing spaces out of string S.
*)
function StrRTrim( S: string ): string; assembler;
asm
  push  ds
  cld
  lds   si, [S]
  mov   cl, ds:[si]
  xor   ch, ch
  jcxz  @@Done
  mov   bx, cx
@@Loop:
  cmp   BYTE PTR ds:[si+bx], ' '
  jne   @@Done
  dec   bx
  loop  @@Loop
@@Done:
  les   di, @Result
  mov   al, cl
  stosb
  inc   si
  rep   movsb
  pop   ds
end;

(*
   Pads the string S with spaces on the left.
*)
function StrLPad( S: string; Len: byte ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  cld
  mov   cl, [Len]
  mov   es:[di], cl                     (* Set length of output string *)
  inc   di
  lodsb
  cmp   al, cl                          (* Need to add spaces? *)
  ja    @@Copy                          (* No, just copy Len bytes *)
  push  ax                              (* Save length of source string *)
  sub   cl, al                          (* Get number of spaces to add *)
  xor   ch, ch
  mov   al, ' '
  rep   stosb
  pop   cx
@@Copy:
  xor   ch, ch
  rep   movsb
  pop   ds
end;

(*
   Pads the string S with spaces on the right.
*)
function StrRPad( S: string; Len: byte ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  cld
  mov   cl, [Len]
  mov   es:[di], cl                     (* Set length of output string *)
  xor   ch, ch
  inc   di
  lodsb
  cmp   al, cl                          (* Need to add spaces? *)
  ja    @@Copy                          (* No, just copy Len bytes *)
  sub   cl, al                          (* Get number of spaces to add *)
  xchg  cl, al
  rep   movsb
  mov   cl, ' '
  xchg  cl, al
  rep   stosb
@@Copy:
  rep   movsb
  pop   ds
end;

(*
   Converts an ASCIIZ string to the Turbo Pascal format.
*)
function StrC2TP( S: pointer ): string; assembler;
asm
  push  ds
  les   di, [S]
  cld
  xor   al, al
  mov   cx, $FFFF
  repne scasb
  inc   cx
  not   cx                              (* Get length of ASCIIZ string *)
  or    ch, ch                          (* Less than 256? *)
  jz    @@Copy                          (* Yes, copy the string *)
  mov   cx, $00FF                       (* Set length to 255 *)
@@Copy:
  lds   si, [S]
  les   di, @Result
  mov   es:[di], cl
  inc   di
  rep   movsb
  pop   ds
end;

(*
   Converts a Turbo Pascal string to the ASCIIZ format.

   Output:
          the pointer D to the buffer that contains the converted string.
*)
function StrTP2C( S: string; D: pointer ): pointer; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, [D]
  mov   dx, di
  cld
  lodsb
  cbw
  xchg  cx, ax
  rep   movsb
  xor   al, al
  stosb
  mov   ax, es
  xchg  ax, dx
  pop   ds
end;

(*
   Copies a block of memory into a Turbo Pascal string.

   Output:
          a string made of N bytes starting at address S.
*)
function Mem2Str( S: pointer; N: byte ): string; assembler;
asm
  push  ds
  lds   si, [S]
  les   di, @Result
  mov   cl, [N]
  xor   ch, ch
  cld
  mov   es:[di], cl
  inc   di
  rep   movsb
  pop   ds
end;

end.
